﻿using System;
using System.Threading;

namespace ConsoleApplication1
{
    class Program
    {
        public int value = 0xff;

        static void LockThread(object lockObject)
        {
            long oldValue = ForceReleaseLock(lockObject);
            lock (lockObject)
            {
                Console.WriteLine("You can't see this message!");
            }
            ForceSetLock(lockObject, oldValue);
        }

        private unsafe static void ForceSetLock(object lockObject, long oldValue)
        {
            if (IntPtr.Size == 4)
            {
                ForceSetLock32(lockObject, (int)oldValue);
            }
            else
            {
                ForceSetLock64(lockObject, oldValue);
            }
        }

        private unsafe static void ForceSetLock32(object lockObject, int oldValue)
        {
            Program pg = lockObject as Program;

            fixed (int* p2 = &pg.value)
            {
                *(p2 - 2) = oldValue;
            }
        }

        private unsafe static void ForceSetLock64(object lockObject, long oldValue)
        {
            Program pg = lockObject as Program;

            fixed (int* p2 = &pg.value)
            {
                long* pValue = (long*)p2;
                *(pValue - 2) = oldValue;
            }
        }

        private unsafe static long ForceReleaseLock(object lockObject)
        {
            return (IntPtr.Size == 4) ? ForceReleaseLock32(lockObject) : ForceReleaseLock64(lockObject);
        }

        private unsafe static long ForceReleaseLock32(object lockObject)
        {
            Program pg = lockObject as Program;
            int lastValue = 0;

            fixed (int* p2 = &pg.value)
            {
                lastValue = *(p2 - 2);
                *(p2 - 2) = 0;
            }

            return lastValue;
        }

        private unsafe static long ForceReleaseLock64(object lockObject)
        {
            Program pg = lockObject as Program;
            long lastValue = 0;

            fixed (int* p2 = &pg.value)
            {
                long* pValue = (long*)p2;
                lastValue = *(pValue - 2);
                *(pValue - 2) = 0;
            }

            return lastValue;
        }

        static unsafe void Main(string[] args)
        {
            Program pg = new Program();

            lock (pg)
            {
                Thread t1 = new Thread(LockThread);
                t1.Start(pg);
                t1.Join();
            }

            int[] test = new int[1];
            test[0] = 0xff;

            Console.WriteLine("\nunlocked int [] ================== ");

            fixed (int* p0 = &test[0])
            {
                WriteArray(p0);
            }

            Console.WriteLine("\nlocked int [] ================== ");

            lock (test)
            {
                fixed (int* p1 = &test[0])
                {
                    WriteArray(p1);
                }
            }

            Console.WriteLine("\nunlocked Program ================== ");
            fixed (int* p2 = &pg.value)
            {
                WriteObject(p2);
            }

            Console.WriteLine("\nlocked Program ================== ");
            lock (pg)
            {
                fixed (int* p3 = &pg.value)
                {
                    WriteObject(p3);
                }
            }

            Thread.Sleep(-1);
        }

        static unsafe void WriteArray(int* pAddress)
        {
            if (IntPtr.Size == 4)
            {
                WriteArray32(pAddress);
            }
            else
            {
                WriteArray64(pAddress);
            }
        }

        static unsafe void WriteArray32(int *pAddress)
        {
            Console.WriteLine("OBJECT Header(SyncBlock Index): " + (*(pAddress - 3)).ToString("x"));
            Console.WriteLine("MethodTable Address: " + (*(pAddress - 2)).ToString("x"));
            Console.WriteLine("배열 요소 크기: " + (*(pAddress - 1)).ToString("x"));
            Console.WriteLine("0번째 배열 값: " + (*(pAddress - 0)).ToString("x"));
        }

        static unsafe void WriteArray64(int* pAddress)
        {
            long* pValue = (long*)pAddress;

            Console.WriteLine("OBJECT Header(SyncBlock Index): " + (*(pValue - 3)).ToString("x"));
            Console.WriteLine("MethodTable Address: " + (*(pValue - 2)).ToString("x"));
            Console.WriteLine("배열 요소 크기: " + (*(pValue - 1)).ToString("x"));
            Console.WriteLine("0번째 배열 값: " + (*(pValue - 0)).ToString("x"));
        }

        static unsafe void WriteObject(int* pAddress)
        {
            if (IntPtr.Size == 4)
            {
                WriteObject32(pAddress);
            }
            else
            {
                WriteObject64(pAddress);
            }
        }

        static unsafe void WriteObject32(int* pAddress)
        {
            Console.WriteLine("OBJECT Header(SyncBlock Index): " + (*(pAddress - 2)).ToString("x"));
            Console.WriteLine("MethodTable Address: " + (*(pAddress - 1)).ToString("x"));
            Console.WriteLine("첫 번째 int 필드 값: " + (*(pAddress - 0)).ToString("x"));
        }

        static unsafe void WriteObject64(int* pAddress)
        {
            long* pValue = (long*)pAddress;

            Console.WriteLine("OBJECT Header(SyncBlock Index): " + (*(pValue - 2)).ToString("x"));
            Console.WriteLine("MethodTable Address: " + (*(pValue - 1)).ToString("x"));

            Console.WriteLine("첫 번째 int 필드 값: " + (*(pAddress - 0)).ToString("x"));
        }
    }
}
